use serde_json::json;

mod test;

fn less_than_num(a: &serde_json::Value, b: &serde_json::Value) -> bool {
    if a.is_f64() {
        let a = a.as_f64().unwrap();
        let b = b.as_f64().unwrap();
        return a < b;
    } else if a.is_i64() {
        let a = a.as_i64().unwrap();
        let b = b.as_i64().unwrap();
        return a < b;
    } else if a.is_u64() {
        let a = a.as_u64().unwrap();
        let b = b.as_u64().unwrap();
        return a < b;
    } else {
        return false;
    }
}

fn greater_than_num(a: &serde_json::Value, b: &serde_json::Value) -> bool {
    if a.is_f64() {
        let a = a.as_f64().unwrap();
        let b = b.as_f64().unwrap();
        return a > b;
    } else if a.is_i64() {
        let a = a.as_i64().unwrap();
        let b = b.as_i64().unwrap();
        return a > b;
    } else if a.is_u64() {
        let a = a.as_u64().unwrap();
        let b = b.as_u64().unwrap();
        return a > b;
    } else {
        return false;
    }
}

pub fn matches(filter: &serde_json::Value, raw_obj: &serde_json::Value) -> bool {
    let filter = filter.as_object().unwrap();
    let obj = raw_obj.as_object().unwrap();

    if filter.len() == 1 {
        let filter_keys: Vec<_> = filter.keys().collect();
        let op = filter_keys.first().unwrap();
        let op_arg = filter.get(op.as_str()).unwrap();
        match op.as_str() {
            "$and" => {
                if let serde_json::Value::Array(and_list) = op_arg {
                    let and_list_bool: Vec<bool> = and_list
                        .iter()
                        .map(|sub_filter| matches(sub_filter, raw_obj))
                        .collect();
                    return !and_list_bool.iter().any(|x| !x);
                } else {
                    return false;
                }
            }
            "$or" => {
                if let serde_json::Value::Array(or_list) = op_arg {
                    let or_list_bool: Vec<bool> = or_list
                        .iter()
                        .map(|sub_filter| matches(sub_filter, raw_obj))
                        .collect();
                    return or_list_bool.iter().any(|x| *x);
                } else {
                    return false;
                }
            }
            _ => {
                if op.starts_with("$") {
                    unimplemented!()
                }
            }
        }
    }

    for (key, val) in filter.iter() {
        if val.is_object() {
            let val_keys: Vec<_> = val.as_object().unwrap().keys().collect();
            if val_keys.first().unwrap().starts_with("$") {
                return match_operator(val, raw_obj, key.as_str());
            } else {
                // nested
                for (_, _) in val.as_object().unwrap() {
                    let new_filter = filter.get(key).unwrap();
                    if let Some(val) = obj.get(key) {
                        return matches(new_filter, val);
                    } else {
                        return false;
                    }
                }
            }
        }

        if let Some(valb) = obj.get(key) {
            if val != valb {
                return false;
            }
        } else {
            return false;
        }
    }

    true
}

fn match_operator(val: &serde_json::Value, raw_obj: &serde_json::Value, key: &str) -> bool {
    let obj = raw_obj.as_object().unwrap();
    let val = val.as_object().unwrap();
    if val.keys().len() == 1 {
        let keys: Vec<_> = val.keys().collect();
        let op = keys.first().unwrap().as_str();
        let op_arg = val.get(op).unwrap();
        match op {
            "$lt" => {
                if let Some(a) = obj.get(key) {
                    return less_than_num(a, op_arg);
                } else {
                    return false;
                }
            }
            "$lte" => {
                if let Some(a) = obj.get(key) {
                    return less_than_num(a, op_arg) || a == op_arg;
                } else {
                    return false;
                }
            }
            "$gt" => {
                if let Some(valb) = obj.get(key) {
                    return greater_than_num(valb, op_arg);
                } else {
                    return false;
                }
            }
            "$gte" => {
                if let Some(a) = obj.get(key) {
                    return greater_than_num(a, op_arg) || a == op_arg;
                } else {
                    return false;
                }
            }
            "$not" => {
                if let Some(inner) = val.get("$not") {
                    if let serde_json::Value::Object(inner) = inner {
                        let new_filter = json!({
                            key: inner
                        });
                        return !matches(&new_filter, raw_obj);
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            "$ne" => {
                if let Some(valb) = obj.get(key) {
                    return valb != op_arg;
                } else {
                    return false;
                }
            }
            "$in" => {
                if let Some(valb) = obj.get(key) {
                    if let serde_json::Value::Array(list) = valb {
                        return list.iter().any(|x| x == op_arg);
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            "$nin" => {
                if let Some(valb) = obj.get(key) {
                    if let serde_json::Value::Array(list) = valb {
                        return !list.iter().any(|x| x == op_arg);
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            "$exists" => {
                if let serde_json::Value::Bool(exists) = op_arg {
                    let valb = obj.get(key).is_some();
                    return *exists == valb;
                } else {
                    return false;
                }
            }
            "$size" => {
                if let Some(serde_json::Value::Array(list)) = obj.get(key) {
                    let val_size = list.len() as u64;
                    if let serde_json::Value::Number(pref_size) = op_arg {
                        let pref_size = pref_size.as_u64().unwrap();
                        return pref_size == val_size;
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            "$regex" => {
                if let serde_json::Value::String(regex_pattern) = op_arg {
                    if let Some(serde_json::Value::String(valb)) = obj.get(key) {
                        let pattern = regex::Regex::new(regex_pattern).unwrap();
                        return pattern.is_match(valb);
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            "$type" => {
                if let Some(valb) = obj.get(key) {
                    if let serde_json::Value::String(type_str) = op_arg {
                        return match type_str.to_lowercase().as_str() {
                            "null" => valb.is_null(),
                            "string" => valb.is_string(),
                            "number" => valb.is_number(),
                            "object" => valb.is_object(),
                            "array" => valb.is_array(),
                            "boolean" => valb.is_boolean(),
                            _ => false,
                        };
                    } else {
                        return false;
                    }
                } else {
                    return false;
                }
            }
            _ => {
                if op.starts_with("$") {
                    unimplemented!()
                }
            }
        }
    }

    return false;
}
